/**********************************************************************
 *
 * PostGIS - Spatial Types for PostgreSQL
 * http://postgis.net
 * Copyright 2011 OSGeo
 *
 * This is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public Licence. See the COPYING file.
 *
 **********************************************************************/

#ifndef PG_LIBCOMMON_H
#define PG_LIBCOMMON_H

#include "lwgeom_pg.h"
#include "lwgeom_transform.h"

#endif /* PG_LIBCOMMON_H */
