/*****************************************************************************
 *
 * This MobilityDB code is provided under The PostgreSQL License.
 * Copyright (c) 2016-2025, Université libre de Bruxelles and MobilityDB
 * contributors
 *
 * MobilityDB includes portions of PostGIS version 3 source code released
 * under the GNU General Public License (GPLv2 or later).
 * Copyright (c) 2001-2025, PostGIS contributors
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written
 * agreement is hereby granted, provided that the above copyright notice and
 * this paragraph and the following two paragraphs appear in all copies.
 *
 * IN NO EVENT SHALL UNIVERSITE LIBRE DE BRUXELLES BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
 * LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION,
 * EVEN IF UNIVERSITE LIBRE DE BRUXELLES HAS BEEN ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 *
 * UNIVERSITE LIBRE DE BRUXELLES SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON
 * AN "AS IS" BASIS, AND UNIVERSITE LIBRE DE BRUXELLES HAS NO OBLIGATIONS TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *****************************************************************************/

/**
 * @brief Aggregate functions for temporal geos.
 */

#ifndef __TGEO_AGGFUNCS_H__
#define __TGEO_AGGFUNCS_H__

/* PostgreSQL */
#include <postgres.h>
/* MEOS */
#include <meos.h>


/*****************************************************************************/

/**
 * Structure storing the SRID and the dimensionality of the temporal geo
 * values for aggregation. Notice that for the moment we do not aggregate
 * temporal geographic points.
 */
struct GeoAggregateState
{
  int32_t srid;
  bool hasz;
};

/*****************************************************************************/

extern bool ensure_geoaggstate(const SkipList *state, int32_t srid, bool hasz);
extern bool ensure_geoaggstate_state(const SkipList *state1,
  const SkipList *state2);

extern Temporal **tpoint_transform_tcentroid(const Temporal *temp, int *count);
extern TSequence *tpointinst_tcentroid_finalfn(TInstant **instants, int count,
  int32_t srid);
extern TSequenceSet *tpointseq_tcentroid_finalfn(TSequence **sequences,
  int count, int32_t srid);
extern Temporal *tpoint_tcentroid_finalfn(SkipList *state);

/*****************************************************************************/

#endif
